#include <algorithm>
#include "common.h"
#include "as.h"
#include "hash.h"
using namespace std; 
//Constructors
AutoSystem::AutoSystem(asNum _as_ID){
	asID = _as_ID;
	packet_counter =0;
	local_clock =0; 
	as_pref_vt.clear();
	
}

AutoSystem::AutoSystem(asNum _as_ID, double lat, double lon){
	as_location.latitude = lat; 
	as_location.longitude =lon; 
	//AutoSystem::AutoSystem(_as_ID);
	asID = _as_ID;
	packet_counter =0;
	local_clock =0; 
	as_pref_vt.clear();
	 
}

/*  Add a prefix to AS (Normally called by Actuator) */
void AutoSystem::addPrefix(Cidr _prefix){
	u32b _idx; 
	//cout << "Adding prefix " <<endl; 
	bool _found = prefix_lookup(_prefix, _idx); 

	if (!_found){ //prefix not foud, simply adding new prefix 
		AS_Prefix_entry_t _tmpASpref;
		_tmpASpref.cdir_prefix = _prefix;
		_tmpASpref.GUID_list.clear();
		as_pref_vt.push_back(_tmpASpref); 
	}
	else{ //Existing, update MASK accordingly 
		if (as_pref_vt[_idx].cdir_prefix.mask_bit >= _prefix.mask_bit) // enlarge
			as_pref_vt[_idx].cdir_prefix.mask_bit = _prefix.mask_bit; 
		else 	shrinkPrefix(_idx, _prefix);  //shrink
	} //end (found) case 
}

/* Shrink prefix 
   Removing all GUID doesn't belong to a new prefix - Pain 
*/
void AutoSystem::shrinkPrefix(u32b prefIndex, Cidr newPref){
	vector<Gentry_t>::iterator _it = as_pref_vt[prefIndex].GUID_list.begin();
	bool partOfPrefix; 
	Hash myHash; 
	while(_it <as_pref_vt[prefIndex].GUID_list.end()){
		partOfPrefix = false; 
		//try all hashing functions 
		for (int hashIndex = 1; hashIndex <K_NUM; hashIndex++){
			u32b hashResult = myHash.hashG2IP((*_it).GUID, hashIndex); 
			if (Common::isPartOfPrefix(newPref,hashResult)){
				partOfPrefix =true; 
				break; 
			}
		}
		if (partOfPrefix) 
			_it++;  //keep it
		else 
			as_pref_vt[prefIndex].GUID_list.erase(_it);   //remove it 
	}
}

/* Update local_clock */
void AutoSystem::update_local_clock(u64b messageTstamp){
	local_clock = max(local_clock,messageTstamp); 
}

/*	add_GUID_entry to specific prefix 
	Add GUID entry to Look up table
	Input: GUID tobe inserted;
		   Timestamp
		   Index of prefix tobe insterted
	Output: 0 if successs; 1 if error occured
	NOTE: 
		+ Overwrite the timestamp if GUID is existed (considered as update) 
		+ Skip verifying AS's responsibility 
		+ Prefix to insert to must be exsit 
*/
void AutoSystem::add_GUID_entry(u64b GUID,u32b prefIndex, u64b timeStamp){
	u32b  _guidIdx; //index of the GUID on the prefix 
	//find it
	bool found = GUID_lookup_with_prefIndx(GUID, prefIndex ,_guidIdx); 

	if (found){ 
		//update clock of that Gentry
		u64b _currTime = as_pref_vt[prefIndex].GUID_list[_guidIdx].updated_time;
		as_pref_vt[prefIndex].GUID_list[_guidIdx].updated_time = max(timeStamp,_currTime); 
		//update AS's clock 
		local_clock = max(timeStamp, local_clock); 
	}
	else{	   
		//find appropriate position then insert the GUID
		Gentry_t _tmpGetry={timeStamp,GUID}; //tobe inserted
		vector<Gentry_t>::iterator _iter;
		for( _iter = as_pref_vt[prefIndex].GUID_list.begin(); 
			_iter < as_pref_vt[prefIndex].GUID_list.end(); 
			_iter++){
				if ( (*_iter).GUID>GUID) 
				break; 
		}
		//at the right position, insert now
		as_pref_vt[prefIndex].GUID_list.insert(_iter,_tmpGetry); 
	}
}


/* GUID lookup within a given prefix index */
bool AutoSystem::GUID_lookup_with_prefIndx(u64b GUID,u32b prefIndx,u32b &GUIDindx){
	Gentry_t tbSearchGen = {0,GUID}; 
	u32b ssize = as_pref_vt[prefIndx].GUID_list.size();
	if (ssize <= 0) {
		GUIDindx = -1;
		return false;
	} 
	int tmp = Common::binary_search(as_pref_vt[prefIndx].GUID_list,0, ssize-1, tbSearchGen);
	if (tmp == -1) return false;
	GUIDindx = (u32b)tmp; 
	return true; 
}
/*	
	Lookup for GUID without knowledge of prefix 
	Input: GUID to be looked up
    Output: Found/not found, prefix Index, and GUID index 
*/
bool AutoSystem::blind_GUID_lookup(u64b GUID,u32b &prefIndex, u32b &GUIDindx){
	Gentry_t tbSearchGen = {0,GUID}; 
	u32b _prefIt = 0; 
	u32b _guidIt; 
	//Go to each prefix
	while (_prefIt < as_pref_vt.size()){
		u32b ssize = as_pref_vt[_prefIt].GUID_list.size(); 
		if (ssize <=0) {
			_prefIt++;
			continue;
		}
		//scan GUID list of that Prefix to find the match
		_guidIt = Common::binary_search(as_pref_vt[_prefIt].GUID_list,0, ssize-1, tbSearchGen); 

		if (_guidIt == -1) _prefIt++; //not found
		else {
			prefIndex = _prefIt;
			GUIDindx = _guidIt; 
			return true; 
		}
	}
	prefIndex = -1; 
	GUIDindx = -1; 
	return false; 		
}
/*	lookup for index of a given prefix 
	Input: GUID to be looked up
    Output: Found (with index of found element) /not found (with -1 as index) 
*/
bool AutoSystem::prefix_lookup(Cidr _prefix,u32b &prefIndex){
	for(u32b i=0; i < as_pref_vt.size(); i++){
		if (as_pref_vt[i].cdir_prefix.prefix == _prefix.prefix){
			prefIndex = i;				//found, return index of the prefix 
			return true; 
		}
	}
	prefIndex =-1;  //If not found then index points to end of vector
	return false; 
}
